(***********************************************************************)
(*                                                                     *)
(*                        OCamlP3l -- Example                          *)
(*                                                                     *)
(*             projet Estime & Cristal, INRIA-Rocquencourt             *)
(*                                                                     *)
(*  Copyright 2003-2006 INRIA.                                         *)
(*  Institut National de Recherche en Informatique et en Automatique.  *)
(*  All rights reserved. This file is distributed only by permission.  *)
(*                                                                     *)
(*  Francois Clement <Francois.Clement@inria.fr>                       *)
(*  Pierre Weis <Pierre.Weis@inria.fr>                                 *)
(*                                                                     *)
(***********************************************************************)

(* $Id: coupling.ml,v 1.1 2007/01/17 23:36:00 fclement Exp $ *)

(******************************************************************************)
(* Program solving Poisson's equation on a rectangle with homogeneous         *)
(* Dirichlet boundary conditions, ie                                          *)
(*          { - Laplacian u = f  on Omega,                                    *)
(*          {             u = 0  on Gamma = boundary of Omega.                *)
(* The Partial Derivative Equation (PDE) is solved using Domain Decomposition *)
(* with overlapping: firstly, on each subdomain, we input some specific       *)
(* Dirichlet conditions, solve the PDE and then output Dirichlet conditions   *)
(* for the neighbor subdomains; secondly, the solution is searched for as the *)
(* fix point of the previously defined Dirichlet to Dirichlet operator.       *)
(* Computation on each subdomain can be run in parallel.                      *)
(*                                                                            *)
(* The OCamlP3l implementation uses the loop skeleton for the fix point       *)
(* and the mapvector skeleton for the concurrent computation on each          *)
(* subdomain.                                                                 *)
(* The poisson's solver is a C program implementing an implicit 5-point       *)
(* scheme which is spawned and connected via stdin/stdout to                  *)
(* computation_on_a_subdomain in a seq skeleton (see "poisson.c").            *)
(* Graphical facilities using gnuplot are provided (see Gnuplot module).      *)
(*                                                                            *)
(* The Rdec2 module provides functions to manipulate regular decompositions   *)
(* of regular rectangular grids.                                              *)
(******************************************************************************)

(******************************************************************************)
(* IMPORTANT:                                                                 *)
(* - remember to fflush data output in the C program, or you will block!      *)
(* - remember not to rely on global references/arrays; thus it is better to   *)
(*   put them inside the data stream.                                         *)
(* - Meta-data, such as the number of workers for farm or mapvector           *)
(*   skeletons, must be set BEFORE the computation network is deployed,       *)
(*   i.e. before any use of skeleton expressions in the pardo application.    *)
(******************************************************************************)

(*************************************************************)
(* Execution is controlled by data read in a config file     *)
(* This file should contain:                                 *)
(* - the name of the file defining the regular decomposition *)
(* - the basename for result files                           *)
(* - the output type: window, ps, none                       *)
(* - the number of iterations for the fix point problem      *)
(* - the number of iterations for the local subproblems      *)
(* - the refreshment step for the graphical device           *)
(* - the number of seconds to sleep before terminating       *)
(*************************************************************)

open Config;;
open Rdec2;;
open Gnuplot;;
open My_data;;
open P3l;;

(* Management of the user options *)

let set_worker_dir, get_worker_dir =
  My_options.saref "worker_dir" "../worker";;
let set_worker_fname, get_worker_fname =
  My_options.saref "worker_fname" "poisson.deb";;
let set_number_of_workers, get_number_of_workers =
  let s, g = My_options.saref "number_of_workers" 0 in
  (fun i ->
    prerr_endline (Printf.sprintf "set number of workers to %d." i); s i),
  (fun () -> prerr_endline "get number of workers."; g ());;
let set_config_dir, get_config_dir =
  My_options.saref "config_dir" "../config_files";;
let set_config_fname, get_config_fname =
  My_options.saref "config_fname" "config_default";;
let set_res_dir, get_res_dir = My_options.saref "res_dir" "res";;

let options = List.rev [
  "--worker_dir", Arg.String set_worker_dir,
  Printf.sprintf
    "<dir>  set <dir> as the directory for the worker program.\
    \n\tDefault file is %S."
    (get_worker_dir ());
  "-w_dir", Arg.String set_worker_dir,
  "<dir>  Same as --worker_dir <dir>.\n";

  "--worker_fname", Arg.String set_worker_fname,
  Printf.sprintf
    "<fname>  set <fname> as the program to be deployed by all workers.\
    \n\tDefault file is %S."
    (get_worker_fname ());
  "-w", Arg.String set_worker_fname,
  "<fname>  Same as --worker_fname <fname>.\n";

  "--number_of_workers", Arg.Int set_number_of_workers,
  Printf.sprintf
    "<nw>  set <nw> as the number of workers.\
    \n\tDefault value is the number of subdomains, nx x ny, \
    \n\tas provided by the Rdec2 configuration file.";
  "-nw", Arg.Int set_number_of_workers,
  "<nw>  Same as --number_of_workers <nw>.\n";

  "--configuration_directory", Arg.String set_config_dir,
  Printf.sprintf
    "<dir>  set <dir> as the configuration directory.\
    \n\tDefault directory is %S."
    (get_config_dir ());
  "-cfg_dir", Arg.String set_config_dir,
  "<dir>  Same as --configuration_directory <dir>.\n";

  "--configuration_file", Arg.String set_config_fname,
  Printf.sprintf
    "<fname>  set <fname> as the configuration file.\
    \n\tDefault file is %S."
    (get_config_fname ());
  "-cfg", Arg.String set_config_fname,
  "<fname>  Same as --configuration_file <fname>.\n";

  "--result_directory", Arg.String set_res_dir,
  Printf.sprintf
    "<dir>  set <dir> as the results directory.\
    \n\tDefault directory is %S."
    (get_res_dir ());
  "-res_dir", Arg.String set_res_dir,
  "<dir>  Same as --result_directory <dir>.\n";
];;

Options.add_all options;;

let usage_msg = Printf.sprintf "\nUsage: %s [options]\n" Sys.argv.(0);;

let parse_cmd_line () = Arg.parse (Options.all ()) (fun _ -> ()) usage_msg;;
parse_cmd_line ();;

(* The elements of the main data stream are of type subdomain array
   (See My_data). *)
let generate_input_stream_item fname cfg rdec =
  let f i bd =
    {config_filename = fname;
     config = cfg;
     rdec = rdec;
     num = i;
     bd = bd;
     iter = 0;
     err = max_float, max_float} in
  let intf0 = make_zero_interface rdec in
  let k = ref 0 in
  (fun () ->
    if !k = 0 then begin incr k; Array.mapi f intf0 end else
    raise P3lstream.End_of_stream);;

let init_poisson (_, num as sd) jtermax (ib, oc) =
  let g = computational_rectangular_grid_of_subdomain sd in
  print_computational_rectangular_grid oc g;
  Printf.fprintf oc "%d\n" jtermax;
  flush oc;;

let computation_on_a_subdomain () =
  let bin = ref None and cout = ref None in
  (fun ({config = cfg; rdec = rd; num = n; bd = bd; iter = i} as subd) ->
    let i = i + 1 in
    let res = Filename.concat (get_res_dir ()) cfg.result_basename in
    let inmax = cfg.inner_itermax in
    let ib, oc =
      My_unix.spawn_init
        (init_poisson (rd, n) inmax) (get_worker_fname ()) bin cout in
    let go =
      match cfg.graphical_output with
      | No -> "."
      | _ ->
          if i mod cfg.display_step <> 0 then "." else
          Printf.sprintf "%s%i" res n in
    Printf.fprintf oc "%s\n" go;
    print_boundary oc bd;
    flush oc;
    let rel_stag, rel_err = Scanf.bscanf ib " %f %f\n" (fun e1 e2 -> e1, e2) in
    let bd = read_boundary ib (size_of_boundary bd) in
    {subd with bd = bd; iter = i; err = rel_stag, rel_err});;

let proj () =
  (fun v ->
    let rdec = get_rdec v in
    let table = connectivity_table_of_regular_decomposition rdec in
    let bds = get_bds v in
    let bds = projection table bds in
    set_bds bds v;
    v);;

let plot () =
  let bin = ref None and cout = ref None in
  (fun v ->
    let cfg = get_config v in
    let go = cfg.graphical_output in
    let bname = Filename.concat (get_res_dir ()) cfg.result_basename in
    begin match go with
    | No -> ()
    | _ ->
      let ic, oc = My_unix.spawn "gnuplot" bin cout in
      let iter = get_iter v in
      if iter mod cfg.display_step = 0 then begin
        let rdec = get_rdec v in
        let nsbd = number_of_subdomains_of_regular_decomposition rdec in
        let gpl = make_plot_command go bname iter nsbd in
        Printf.fprintf oc "%s\n" gpl;
        flush oc end; end;
    v);;

let convergence =
  let t0 = ref (Unix.gettimeofday ()) in
  (fun v ->
    let iter = get_iter v in
    let f (rs1, re1) {err = rs2, re2} = max rs1 rs2, max re1 re2 in
    let rel_stag, rel_err = Array.fold_left f (0. , 0.) v in
    let dt = Unix.gettimeofday () -. !t0 in
    Printf.eprintf
      "iter = %2d -> rel_stag = %5.0e, rel_err = %6.2f%% (time = %.2f s)\n"
      iter rel_stag (100. *. rel_err) dt;
    flush stderr;
    let cfg = get_config v in
    iter < cfg.outer_itermax);;

let print_result v =
  let fname = get_config_filename v in
  Printf.eprintf
    "\nComputation from configuration file %S completed.\n\n" fname;
  let cfg = get_config v in
  match cfg.graphical_output with
  | Window -> Unix.sleep cfg.a_while
  | _ -> ();;

let get_global_variables () =
  let cfg_fname = Filename.concat (get_config_dir ()) (get_config_fname ()) in
  let cfg = parse_configuration_file cfg_fname in
  let rdec_fname =
    Filename.concat (get_config_dir ()) cfg.rdec_config_filename in
  let rdec = parse_rdec_config_file rdec_fname in
  check_regular_decomposition rdec;
  check_graphical_output cfg.graphical_output;
  if get_number_of_workers () = 0 then begin
    let nsbdx, nsbdy = number_of_subdomains_of_regular_decomposition rdec in
    set_number_of_workers (nsbdx * nsbdy) end;
  cfg_fname, cfg, rdec;;
let cfg_fname, cfg, rdec = get_global_variables ();;

let exprbody () =
  loop (convergence,
        mapvector (seq computation_on_a_subdomain, get_number_of_workers ()) |||
        seq proj |||
        seq plot);;
let expr_skl = parfun exprbody;;

pardo (fun () ->
(*  let cfg_fname, cfg, rdec = get_global_variables () in*)
  P3lstream.iter
    print_result
    (expr_skl
       (P3lstream.of_fun
          (generate_input_stream_item cfg_fname cfg rdec))));;
